//Copyright (c) 1999, Art Gittleman
//This example is provided WITHOUT ANY WARRANTY either expressed or implied.

/* Uses wait and notify to enable
 * producer and consumer threads to
 * cooperate in using a buffer.
 */

// Updated to concurrent by Dr. Spiegel
// Updated to use semaphores
//   Synchronized removed from Buffer methods; semaphore handles this
  
public class ConcurrentPutGet {

  class Buffer {
    int buffer[];             
    int size;        // buffer size
    int putpos=0;
    int getpos=0;
    int number=0;
    Semaphore ProducerMutex,ConsumerMutex, Mutex; 
    Semaphore Full,Empty;	// Counting Semaphores
    
    // Synchronized methods in sempaphore class now

    public Buffer(int size)
    { this.size=size;
      buffer = new int [size]; 
      // Binary Semaphores
      ProducerMutex=new Semaphore(1); 	
      ConsumerMutex=new Semaphore(1); 	
      Mutex=new Semaphore(1); 	
      Full=new Semaphore(size);
      Empty=new Semaphore(0);
    }

    public void put(int value) throws InterruptedException {
      ProducerMutex.waitSemaphore();
      Full.waitSemaphore();  // Wait if full buffer
      // Number and the buffer are shared between producers and consumers
      Mutex.waitSemaphore();
      number++;
      buffer[putpos] = value;
      System.out.println("Put "+value+" Number:"+number);
      Mutex.signalSemaphore();
      Empty.signalSemaphore();
      // putpos is shared only by consumers; why must the wait be on top???
      putpos=(putpos+1)%size;
      ProducerMutex.signalSemaphore();
    }

    public int get() throws InterruptedException {
      ConsumerMutex.waitSemaphore();
      Empty.waitSemaphore();
      Mutex.waitSemaphore();
      number--;
      int n = buffer[getpos];
      System.out.println("Get "+n+" Number:"+number);
      Mutex.signalSemaphore();
      Full.signalSemaphore();
      // getpos is shared only by consumers
      getpos=(getpos+1)%size;
      ConsumerMutex.signalSemaphore();
      return n;
    }
  }

  class Producer extends Thread {
     Buffer buf;

     Producer(Buffer b) { 
        buf = b;
     }			

     public void run() {
       for(int i=1; ; i++)
         try {
             buf.put(i);
	     sleep(100);
           }catch (InterruptedException e){
              e.printStackTrace();
           }
     }
  }

  class Consumer extends Thread {
     Buffer buf;

     Consumer(Buffer b) { 
       buf = b;
     }

     public void run() {
//       for (int i=1; i<=10;i++)
      while (true)
         try {
           buf.get();
	   sleep(100);
         }catch (InterruptedException e) {
               e.printStackTrace();
         }
     }
  }

  public static void main(String[] args) {
    if (args.length!=3) {
	System.out.println("Usage: PutGet  <Buffer Size> <#Producers>  <#Consumers>");
        System.exit(0);
    }
    ConcurrentPutGet pg = new ConcurrentPutGet();
    Buffer b = pg.new Buffer(Integer.parseInt(args[0]));
    Producer p[] = new Producer[Integer.parseInt(args[1])];
    for (int pIdx=0;pIdx<Integer.parseInt(args[1]);pIdx++) {
      p[pIdx]=pg.new Producer(b);
      p[pIdx].start();
    }
    Consumer c[] = new Consumer[Integer.parseInt(args[2])];
    for (int cIdx=0;cIdx<Integer.parseInt(args[2]);cIdx++) {
      c[cIdx]=pg.new Consumer(b);
      c[cIdx].start();
    }
  }
}

